#include <grwin.h>
#include <windows.h>
#include <stdio.h>
#include <process.h>
#include "thread.h"
#include "mList.h"
#include "mandelcommon.h"

#define MAX_CPU_CORE 32
#define MIN_CPU_CORE 1

static HANDLE listMutex;
static HANDLE threadSemaphore;
#ifdef DEBUG
static HANDLE DEBUG_printMutex;
#endif

static void threadReleaseMutex(HANDLE mutexHandle);
static int threadGetMutex(HANDLE mutexHandle,int waitMilliSec);
static void threadCalcTask(pCalcTask taskInfo);
static void thread(pMDesc taskQ);

#ifdef DEBUG
void DEBUG_showReceivedTask(pCalcTask task);
#endif

/**********************************
 CPUの論理コア数を取得
 return:
 論理コア数
**********************************/
int threadGetLogicalProcessorNum(void){
  SYSTEM_INFO sysInfo;
  int ret;
  
  GetSystemInfo(&sysInfo);

  if(sysInfo.dwNumberOfProcessors > MAX_CPU_CORE){
    ret = MAX_CPU_CORE;
  }else if(sysInfo.dwNumberOfProcessors < MIN_CPU_CORE){
    ret = MIN_CPU_CORE;
  }else{
    ret = sysInfo.dwNumberOfProcessors;
  }
  return ret;
}

/**************************************
 排他-ミューテックスを作成
 return:
 ミューテックスのハンドル
 *************************************/
HANDLE threadCreateMutex(void){
  return CreateMutex(NULL,FALSE,NULL);
}

/*************************************
 ミューテックスの削除
 return:
 無し
 ************************************/
void threadCloseHandle(HANDLE mutexHandle){
  if(CloseHandle(mutexHandle) == 0){
    fprintf(stderr,"Failed Closing Mutex\n");
  }
  return;
}

/*************************************
 ミューテックスの解放
 return:
 無し
 ************************************/
static void threadReleaseMutex(HANDLE mutexHandle){
  if(ReleaseMutex(mutexHandle) == 0){
    fprintf(stderr,"Failed Releasing Mutex\n");
  }
  return;
}

/**************************************
 ミューテックスの取得
 waitMilliSecは待たない場合0
 待ち続けるときはINFINITE
 return:
 成功: 0
 失敗: -1
 *************************************/
static int threadGetMutex(HANDLE mutexHandle,int waitMilliSec){
  int ret;

  if(WaitForSingleObject(mutexHandle,waitMilliSec) == WAIT_OBJECT_0){
    ret = 0;
  }else{
    ret = -1;
  }
  
  return ret;  
}
  
/**************************************
 各画素について反復回数を求めて記録する
 return:
 無し
 *************************************/
static void threadCalcTask(pCalcTask task){
#ifdef USE_FLOAT128
  __float128 x,y,nX,nY,preX;
#else
  double x,y,nX,nY,preX;
#endif
  int i,j,k,line,iterate;
  int tmpLine[PIXEL_X];

  x = task->x;
  y = task->y;
  line = task->line;
  iterate = task->iterate;

  for(i = 0;i < line;i++){
    for(j = 0;j < PIXEL_X; j++){
      nX = x;
      nY = y;
      for(k = 0;k < iterate;k++){
	if(nX*nX + nY*nY > 4.0){
	  //task->brot[i * PIXEL_X + j] = k;
	  tmpLine[j] = k;
	  //InterlockedExchange((LONG *)&(task->brot[i*PIXEL_X + j]),k);
	  break;
	}
	preX = nX;
	nX = nX*nX - nY*nY + x;
	nY = preX*nY*2 + y;
      }
      if(k == iterate){
	//task->brot[i * PIXEL_X + j] = k - 1;
	tmpLine[j] = k - 1;
	//InterlockedExchange((LONG *)&(task->brot[i*PIXEL_X + j]),k-1);
      }
      x += task->delta;
      //WaitForSingleObject(DEBUG_printMutex,INFINITE);
      //ReleaseMutex(DEBUG_printMutex);
    }
    memcpy(task->brot + i * PIXEL_X,tmpLine,sizeof(tmpLine));
    
    x = task->x;
    y -= task->delta;
    task->pRep->haveDone++;
#ifdef DEBUG
    Sleep(10);
#endif
  }
#ifdef DEBUG
  fprintf(stderr,"calcTask end\n");
#endif
  return;
}

/**************************************
 同期-セマフォを作成
 return:
 handle of Semephore
 *************************************/
HANDLE threadCreateSemaphore(int maxResource){
  return CreateSemaphore(NULL,0,maxResource,NULL);
}

/**************************************
 スレッドで使用するセマフォ・ミューテックスの準備と
 スレッドの生成
 return:
 成功: 0
 失敗: -1
 *************************************/
#define STACK_SIZE 0x400
int threadInit(int threadNum,pMDesc taskQ){
  int i;
  DWORD threadId;

  if((threadSemaphore = threadCreateSemaphore(threadNum)) == NULL){
    return -1;
  }
  if((listMutex = threadCreateMutex()) == NULL){
    CloseHandle(threadSemaphore);
    return -1;
  }
#ifdef DEBUG
  if((DEBUG_printMutex = threadCreateMutex()) == NULL){
    CloseHandle(threadSemaphore);
    CloseHandle(listMutex);
    return -1;
  }
#endif
  
  for(i = 0;i < threadNum;i++){
    
    _beginthreadex(NULL,STACK_SIZE,(void *)thread,(void *)taskQ,0,NULL);
    /*
    if(CreateThread(NULL,STACK_SIZE,(void *)thread,(void *)taskQ,0,&threadId) == NULL){
      fprintf(stderr,"failed to create thread\n");

      CloseHandle(threadSemaphore);
      CloseHandle(listMutex);
      return -1;
    }
    */
  }

  return 0;  
}

static void thread(pMDesc taskQ){
  calcTask txTsk,*rvTsk;
  pTaskNode node;
  
  while(1){
    WaitForSingleObject(threadSemaphore,INFINITE); /* 同期 */
    
    /* 排他 */
    WaitForSingleObject(listMutex,INFINITE);
    node = (pTaskNode)mListGet(taskQ);
    ReleaseMutex(listMutex);
    /* end of critical section */
    if(node != NULL){
      rvTsk = node->pTask;
      free(node);
    }
    if(rvTsk == NULL){
#ifdef DEBUG
      fprintf(stderr,"thread> get NULL task\n");
#endif
      continue;
    }
    memcpy(&txTsk,rvTsk,sizeof(calcTask));
    free(rvTsk);
#ifdef DEBUG
    //    DEBUG_showReceivedTask(&txTsk);
#endif
    threadCalcTask(&txTsk);
  }

  return;
}

int startThread(int threadNum){
  int ret;
  ret = ReleaseSemaphore(threadSemaphore,threadNum,NULL);
  if(ret != 0){
    return 0;
  }
  return -1;
}

#ifdef DEBUG
void DEBUG_showReceivedTask(pCalcTask task){
  WaitForSingleObject(DEBUG_printMutex,INFINITE);
  printf("thread> Received Task ***************\n");
  printf("x=%f, y=%f, delta=%f, line=%d, brot=0x%p, iterate=%d\n",
	 task->x,task->y,task->delta,task->line,task->brot,task->iterate);
  ReleaseMutex(DEBUG_printMutex);
  return;
}
#endif
